import sqlite3 as sql
import numpy as np
from . import TCP
import time

indexColumnName = "idx"
timeColumnName = "timestamp"



class Database():
    def __init__(self):
        self.path = ""
        self.opened = False
        
    def execute(self, data : str):
        if(not self.opened):
            self.open()
        assert self.opened, "Cannot open database"
        return self.db.execute(data)
        
    def commit(self, alias : str = None, full : bool = False):
        self.db.commit()
        if(not alias is None):
            TCP.updateDatabase(alias, full)
        else:
            TCP.updateDatabase()
        
        
    def open(self):
        self.path = TCP.getDatabasePath()
        if(type(self.path) == bytearray):
            self.path = self.path.decode()
        try:
            self.db = sql.connect(self.path, timeout=10)
            self.opened = True
        except sql.Error as er:
            print("Couldn't open data base : " + er.args)
            self.opened = False

    def _blockExists(self, alias):
        val = self.execute('SELECT count(*) FROM sqlite_master WHERE type=\'table\' AND name=\'' + alias + '\'')
        exists = False
        for v in val:
            exists = v[0] > 0
        return exists
        
    def _listBlocks(self):
        val = self.execute('SELECT name FROM sqlite_master WHERE type=\'table\'')
        tables = []
        for x in val:
            name = x[0]
            if(name != 'sqlite_sequence'):
                tables.append(name)
        return tables
    
    
    def _newBlock(self, alias : str):
        if(not self._blockExists(alias)):
            command = 'CREATE TABLE \'' + alias + '\' (\"' + indexColumnName + '\" INTEGER PRIMARY KEY AUTOINCREMENT, \"' + timeColumnName + '\" REAL)'
            self.execute(command)
            #for c in columns:
            #    if(c[1] == int):
            #        columnsStr += c[0] + " INTEGER"
            #    elif(c[1] == float):
            #        columnsStr += c[0] + " REAL"
            #    elif(c[1] == str):
            #        columnsStr += c[0] + " TEXT"
            #    elif(c[1] == bytearray):
            #        columnsStr += c[0] + " BLOB"
            #    count += 1
        self.commit()   
    
    def _removeBlock(self, alias : str):
        self.execute('DROP TABLE ' + alias)
        self.execute('DELETE FROM sqlite_sequence WHERE name = \'' + alias + '\'')
        self.commit()
    
    def _truncate(self, alias):
        self.execute('DELETE FROM ' + alias)
        self.execute('DELETE FROM sqlite_sequence WHERE name = \'' + alias + '\'')
        self.commit(alias, True)
            
db = Database()

    
def removeBlock(alias : str):
    if(db._blockExists(alias)):
        db._removeBlock(alias)
        return True
    else:
        return False
    
    
    
def block(alias : str):
    if(not db._blockExists(alias)):
        db._newBlock(alias)
    return Block(alias)
    
    
    
def listBlocks():
    return db._listBlocks()
    


class Block:
    def __init__(self, alias : str):
        self.alias = alias
        self.exists = True

    def _listColumns(self):
        if(not self.exists):
            return None
        else:
            cursor = db.execute('select * from ' + self.alias)
            names = list(map(lambda x: x[0], cursor.description))
            return names
    
    def _columnName(self, index : int):
        
        if(index == 0):
            # "index" column
            name = indexColumnName
        elif(index == 1):
            # "time" column
            name = timeColumnName
        else:
            name = ""
            value = index - 1
            while(value > 0):
                mod = (value-1) % 26
                c = chr(mod + ord('A'))
                name = c + name
                value = (value-mod)//26
        return name
        
    def columnTypeByDataType(self, dataType):
        if(dataType is None):
            return "NULL"
        if(dataType == int):
            return "INTEGER"
        elif(dataType == float):
            return "REAL"
        elif(dataType == str):
            return "TEXT"
        elif(dataType == bytearray):
            return "BLOB"
        else:
            return "NULL"

    def _rows(self):
        vals = db.execute('SELECT COUNT(*) FROM ' + self.alias + '')
        for v in vals:
            return v[0]

    def _insertRow(self, timestamp : float):
        command = 'INSERT INTO ' + self.alias + ' (%s) VALUES(%f)' % (timeColumnName, timestamp)
        try:
            db.execute(command)
        except:
            print("Error trying to do command : " + command)
        return self._rows()# Returning the index of the added row
        
    def _setSingleValue(self, line, column, value):
        # Line=0 -> column A
        assert line <= self._rows(), "Current line doesn't exist !"

        strVal = str(value)
        strVal = strVal.replace('\'', '\'\'')
        
        command = 'UPDATE %s SET %s = \'' % (self.alias, self._columnName(column+2)) + strVal + '\' WHERE ' + indexColumnName + '= %d' % (line)
        try:
            db.execute(command)      
        except:
            print("Error trying to do command : " + command)

    def _addColumn(self, alias : str, columnType : str):
        db.execute('ALTER TABLE \'' + self.alias + '\' ADD COLUMN \'' + alias + '\' ' + columnType + ' ;')
        
    
    def log(self, values):
        if(not self.exists):
            print("Block has been deleted")
            return None
        else:
            self.logAt(time.time(), values)
        
    def logAt(self, timestamp : float, values):
        if(not self.exists):
            print("Block has been deleted")
        else:
            columnCount = len(self._listColumns())

            if(type(values) == tuple or type(values) == list):
                # List / Tuple
                columnsNeeded = len(values)
            elif(type(values) == np.ndarray):
                # List of values
                columnsNeeded = values.size
            else:
                # Single value
                columnsNeeded = 1

            # Adding missing columns
            for i in range(0, columnsNeeded+2): #2 : index + time
                if(i >= columnCount):
                    self._addColumn(self._columnName(i), self.columnTypeByDataType('TEXT'))
            
            
            # Adding a new row with the timestamp
            row = self._insertRow(timestamp)
            if(columnsNeeded > 1):
                #Adding multiple values
                for i in range(0, len(values)):
                    self._setSingleValue(row, i, values[i]) 
            else:
                #Adding a single value
                self._setSingleValue(row, 0, values)

            db.commit(self.alias, False)

    def clear(self):
        if(not self.exists):
            print("Block has been deleted")
        else:
            db._truncate(self.alias)

    def delete(self):
        db._removeBlock(self.alias)
        self.exists = False
        



from enum import Enum
import socket
import struct
import os


class TCPPacketsTypes(int, Enum):
    # Error packet
    ERROR = 0
    # Generic acknowledge
    GENERIC_OK = 1
    # From Scripts/scriptserver.h  ! WARNING ! : IN <-> OUT changed
    OUT_WRITE_DEVICE = 2
    IN_WRITE_DEVICE_STATUS = 3
    OUT_READ_DEVICE = 4
    IN_READ_DEVICE_STATUS = 5
    IN_READ_DEVICE_DATA = 6
    # Buffer commands
    OUT_READ_BUFFER = 7
    IN_BUFFER_DATA = 8
    # Blocks commands
    OUT_DATABASE_PATH_REQUEST = 9
    IN_DATABASE_PATH = 10
    # Driver request
    OUT_DRIVER_PATH = 11
    IN_DRIVER_PATH = 12
    # Blocks updating
    OUT_UPDATE_BLOCKS = 13
    # Closing device
    OUT_CLOSE_DEVICE = 14


class ErrorCode(int, Enum):
    INVALID_ALIAS = 0  # Request for the wrong alias
    WRITE_ERROR = 1


errorMessages = {
    ErrorCode.INVALID_ALIAS: "Alias invalide",
    ErrorCode.WRITE_ERROR: "Erreur lors de l'ecriture"}


LOCALHOST = "127.0.0.1"
DEFAULT_PORT = 8445
DEFAULT_TIMEOUT_MS = 100000


_opened = False


def close():
    __s__.close()


def open(port=DEFAULT_PORT):
    # print("Opening...")
    global __s__
    global _opened
    retry = 2
    while(retry):
        retry -= 1
        try:
            __s__ = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            __s__.settimeout(DEFAULT_TIMEOUT_MS / 1000)
            __s__.connect((LOCALHOST, port))
            _opened = True
        except socket.error as exc:
            print("closing...")
            close()
            # Retry once
            _opened = False
            if(retry == 0):  # After second try
                print("second try fail")
                raise Exception("Connexion à l'hôte impossible : %s" % exc)


def sendPacket(data: bytearray):
    #print("Sending packet...")
    global __s__
    global _opened
    if(_opened == False):
        open()

    if(_opened):
        try:
            __s__.send(data)
        except Exception:
            # recreate the socket and try again
            close()
            open()  # Add the port here !
            __s__.send(data)


def readPacket(timeout_ms=0):
    global __s__
    global _opened
    if(timeout_ms > 0):
        # +1 to let the communication happend (and C++ to do its thing)
        __s__.settimeout(timeout_ms / 1000 + 1)
    else:
        __s__.settimeout(DEFAULT_TIMEOUT_MS / 1000)
    packetType = 0
    data = b''
    if(not _opened):
        open()
    if(_opened):
        maxBytes = 65532
        length = maxBytes
        raw = b''
        while length == maxBytes:
            buf, _ = __s__.recvfrom(maxBytes)
            length = len(buf)
            raw += buf
        packetType = raw[0]
        data = raw[1:]

    return packetType, data


def printErrorCode(code: int):
    err = errorMessages[ErrorCode(code)]
    if(err is None):
        print("Erreur inconnue")
    else:
        print(err)


def writeToDevice(alias: str, data: bytearray):
    packet = struct.pack('!BB%dsh%ds' % (len(alias), len(data)),
                         TCPPacketsTypes.OUT_WRITE_DEVICE.value, len(alias), alias.encode(), len(data), data)

    sendPacket(packet)
    type, data = readPacket()
    if(type == TCPPacketsTypes.ERROR.value):
        printErrorCode(data[0])
        return False
    else:
        return True


def readFromDevice(alias: str, storeInBuffer: bool = False, timeout_ms=None):
    if(timeout_ms is None):
        timeout_ms = DEFAULT_TIMEOUT_MS
    elif(timeout_ms > 0xFFFFFFFF or timeout_ms < 0):
        timeout_ms = DEFAULT_TIMEOUT_MS

    packet = struct.pack('!BB%dsBL' % len(alias),
                         TCPPacketsTypes.OUT_READ_DEVICE.value, len(alias), alias.encode(), storeInBuffer, timeout_ms)

    sendPacket(packet)
    if(storeInBuffer):  # Will probably get deleted, the buffer isn't used
        _, status = readPacket(timeout_ms)
        if(status == 0):
            print("Erreur lors de l'envoi")
    else:
        type, data = readPacket(timeout_ms)
        if(type == TCPPacketsTypes.ERROR.value):
            printErrorCode(data[0])
        else:
            return data[4:]


def getDatabasePath():
    packet = struct.pack('!B', TCPPacketsTypes.OUT_DATABASE_PATH_REQUEST.value)
    sendPacket(packet)
    _, path = readPacket()
    return path[2:]


def getDeviceDriverPath(alias: str):
    packet = struct.pack('!BB%ds' % len(
        alias), TCPPacketsTypes.OUT_DRIVER_PATH, len(alias), alias.encode())
    sendPacket(packet)
    type, data = readPacket()
    if(type == TCPPacketsTypes.ERROR.value):
        printErrorCode(data[0])
        return None
    else:
        driverName = data[2:]
        return driverName.decode()


def updateDatabase(alias: str = None, full: bool = False):
    if(alias is None):
        packet = struct.pack('!BB', TCPPacketsTypes.OUT_UPDATE_BLOCKS, 0)
    else:
        packet = struct.pack('!BB%dsB' % len(
            alias), TCPPacketsTypes.OUT_UPDATE_BLOCKS, len(alias), alias.encode(), full)

    sendPacket(packet)


def closeDevice(alias):
    packet = struct.pack('!BB%ds' % len(
        alias), TCPPacketsTypes.OUT_CLOSE_DEVICE.value, len(alias), alias.encode())
    sendPacket(packet)
    type, data = readPacket()
    if(type == TCPPacketsTypes.ERROR.value):
        printErrorCode(data[0])
        return False
    else:
        return True



from . import TCP
import os
import importlib


def device(alias: str):
    # full driver path
    driverPath = TCP.getDeviceDriverPath(alias)
    if(driverPath is not None):
        # extract the filename without the extension
        path, module = os.path.split(driverPath)
        moduleName, extension = os.path.splitext(module)

        spec = importlib.util.spec_from_file_location(moduleName, driverPath)
        if(spec is None):
            print(f"Le driver n'as pas été trouvé ({driverPath})")
            return None
        else:
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
            return module.new(Device(alias))


class Device():
    def __init__(self, alias: str):
        self.alias = alias

    def write(self, data):
        if(type(data) == str):
            data = data.encode()
        return TCP.writeToDevice(self.alias, data)

    def read(self, timeout_ms=0):
        return TCP.readFromDevice(self.alias, False, timeout_ms)

    def close(self):
        return TCP.closeDevice(self.alias)



from .blocks import block

consoleBlock = block('console')

def _log(message : str, level : str):
    consoleBlock.log([level, message])

def clear():
    consoleBlock.clear()

def logInfo(message : str):
    _log(message, 'info')
    
def logWarning(message : str):
    _log(message, 'warning')

def logError(message : str):
    _log(message, 'error')
    
def logDebug(message : str):
    _log(message, 'debug')
    
def logCritical(message : str):
    _log(message, 'critical')
    
def logTrace(message : str):
    _log(message, 'trace')
    
def logUnknown(message : str):
    _log(message, 'unknown')




from setuptools import setup, find_packages



# Thanks to this guy who helped me a lot for writing this package
# https://www.youtube.com/watch?v=GIF3LaRqgXo

setup(
    name='pyatmp',
    version='0.9.0',
    description='Package for interfacing with ATMP software',
    author='Sébastien Deriaz',
    author_email='sebastien.deriaz1@gmail.com',
    #url='https://github.com/SebastienDeriaz/pyatmp',
    package_dir={'' : 'src'},
    #license='...'
    packages=find_packages(where="src"),
    #install_requires=['enum', 'socket', 'struct', 'importlib', 'sqlite3', 'numpy', 'time']
)